/*
 * animatch.c: animate images together.
 *
 * Copyright 2002 by Akkana Peck.
 * You are free to use or modify this code under the Gnu Public License.
 */

#include <stdlib.h>       // for getenv()
#include <stdio.h>

#include <math.h>

//#include <X11/Xlib.h>

#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk-pixbuf/gdk-pixbuf-xlib.h>

#include <X11/keysym.h>

static Display *dpy;
static int screen;
static Window win;
static GC gc;

static int XSize = 500;
static int YSize = 300;
static GdkPixbuf* image;

static int ArgC, ArgP;
static char** ArgV;

static long eventMask = ( ExposureMask
                          | ButtonPressMask| ButtonReleaseMask
                          | KeyPressMask
                          | StructureNotifyMask );

typedef struct XY_s
{
    int x, y;
} XY;
#define NPOINTS 20
static XY Points[NPOINTS];
static XY FirstPoints[NPOINTS];
static int NumPoints = 0;
static int NumFirstPoints = 0;
static XY FirstSize;

void PrintPoints()
{
#ifdef DEBUG
    int i;
    printf("%s (%d x %d):", ArgV[ArgP], XSize, YSize);
    for (i=0; i<NumPoints; ++i)
        printf(" (%d, %d)", Points[i].x, Points[i].y);
    printf("\n");
#endif

    if (ArgP > 1 && NumPoints >= 2 && NumFirstPoints >= 2)
    {
        // Calculate how to transform this image to match the first.
        // For now, we only take into account the first two points.
        // First, calculate rotation.
        int dy0 = FirstPoints[1].y - FirstPoints[0].y;
        int dy1 = Points[1].y - Points[0].y;
        int dx0 = FirstPoints[1].x - FirstPoints[0].x;
        int dx1 = Points[1].x - Points[0].x;
        double angle = atan2(dy1, dx1) - atan2(dy0, dx0);
        double absangle = fabs(angle);
        double r, theta;
        double xx, yy;
        double newwidth, newheight;
        if (angle > M_PI) angle -= M_PI;

        // If we rotated, then the image size has increased,
        // so figure out by how much:
        newwidth = XSize * cos(absangle) + YSize * sin(absangle);
        newheight = XSize * sin(absangle) + YSize * cos(absangle);
#ifdef DEBUG
        printf("newsize = %lf x %lf\n", newwidth, newheight);
        printf("        = %d * %lf + %d * %lf, %d * %lf + %d * %lf\n",
               XSize, cos(angle), YSize, sin(angle),
               XSize, sin(angle), YSize, cos(angle));
#endif

        // Now, figure out how to translate the coordinates
        // after a rotation around the center.
        // First, translate the two current points to radial:
#ifdef DEBUG
        printf("Point: (%d, %d)\n", Points[0].x, Points[0].y);
#endif
        xx = Points[0].x - XSize/2;
        yy = Points[0].y - YSize/2;
#ifdef DEBUG
        printf("Translate: (%lf, %lf)\n", xx, yy);
#endif
        r = sqrt(xx*xx + yy*yy);
        theta = atan2(-yy, xx) + angle;   // theta after rotation
#ifdef DEBUG
        printf("r = %lf, theta = %lf\n", r, theta*180/M_PI);
#endif
        // Now translate back:
        xx = r * cos(theta) + newwidth/2;
        yy = -r * sin(theta) + newheight/2;
#ifdef DEBUG
        printf("Translate back: (%lf, %lf) -> (%lf, %lf)\n",
               xx - newwidth/2, yy-newheight/2, xx, yy);
#endif
        // xx and yy now are the coordinates of the zeroth point
        // in the coordinate system of the rotated and hence bigger image.
        dx0 = xx - FirstPoints[0].x;
        dy0 = yy - FirstPoints[0].y;
        //printf("\tCut off the first (%d, %d)\n", dx0, dy0);

        printf("pnmrotate %lf %s > foo.pnm\n", angle * 180. / M_PI, ArgV[ArgP]);

        // Add any necessary padding:
        if (dx0 < 0 && dy0 < 0)
        {
            printf("pnmpad -black -l%d -t%d foo.pnm > bar.pnm\n", -dx0, -dy0);
            dx0 = 0;  dy0 = 0;
        }
        else if (dx0 < 0)
        {
            printf("pnmpad -black -l%d foo.pnm > bar.pnm\n", -dx0);
            dx0 = 0;
        }
        else if (dy0 < 0)
        {
            printf("pnmpad -black -t%d foo.pnm > bar.pnm\n", -dy0);
            dy0 = 0;
        }
        else printf("mv foo.pnm bar.pnm\n");
                   
        // dx0, dy0 are the offset of the upper left edge (after padding).
        // Calculate dx1, dy1 as the last coordinate of the lower right:
        dx1 = dx0 + FirstSize.x;
        dy1 = dy0 + FirstSize.y;
        if (dx1 > XSize && dy1 > YSize)
        {
            printf("pnmpad -black -r%d -b%d bar.pnm > foo.pnm\n",
                   dx1 - XSize, dy1 - YSize);
        }
        else if (dx1 > XSize)
        {
            printf("pnmpad -black -r%d bar.pnm > foo.pnm\n", dx1 - XSize);
        }
        else if (dy1 > YSize)
        {
            printf("pnmpad -black -b%d bar.pnm > foo.pnm\n", dy1 - YSize);
        }

        printf("pnmcut -left %d -top %d -right %d -bottom %d foo.pnm > new/%s\n",
               dx0, dy0, dx1, dy1, ArgV[ArgP]);
    }
    else printf("cp %s new/%s\n", ArgV[ArgP], ArgV[ArgP]);
    printf("\n");

    NumPoints = 0;
}

void ShowImage()
{
    image = gdk_pixbuf_new_from_file(ArgV[ArgP]);
    if (!image)
    {
        fprintf(stderr, "%s: Couldn't open file\n", ArgV[ArgP]);
        return;
    }

    XSize = gdk_pixbuf_get_width(image);
    YSize = gdk_pixbuf_get_height(image);
    if (ArgP == 1) {
        FirstSize.x = XSize;
        FirstSize.y = YSize;
    }

    // Turn off event processing, else we get an infinite Expose loop:
    XResizeWindow(dpy, win, XSize, YSize);
    gdk_pixbuf_xlib_render_to_drawable(image, win, gc,
                                       0, 0, 0, 0, XSize, YSize,
                                       XLIB_RGB_DITHER_NONE, 0, 0);
    //DisplayImages(imageInfo, image);
    XSync(dpy, 1);
}

int HandleEvent()
{
    XEvent event;
    char buffer[20];
    KeySym keysym;
    XComposeStatus compose;

    XNextEvent(dpy, &event);
    switch (event.type)
    {
      case Expose:
          //printf("Expose\n");
          ShowImage();
          break;
      case ConfigureNotify:
          XSize = event.xconfigure.width;
          YSize = event.xconfigure.height;
          //printf("ConfigureNotify: now (%d, %d)\n", XSize, YSize);
          //XSelectInput(dpy, win, eventMask);
          break;
      case ButtonPress:
          Points[NumPoints].x = event.xbutton.x;
          Points[NumPoints].y = event.xbutton.y;
          if (ArgP == 1)
          {
              FirstPoints[NumFirstPoints].x = Points[NumPoints].x;
              FirstPoints[NumFirstPoints].y = Points[NumPoints].y;
              ++NumFirstPoints;
          }
          ++NumPoints;
          break;
      case ButtonRelease:
          break;
      case KeyPress:
          XLookupString(&(event.xkey), buffer, sizeof buffer,
                        &keysym, &compose);
          switch (keysym)
          {
            case XK_space:
                // Go to next image
                PrintPoints();
                if (ArgP >= ArgC-1)
                    return -1;
                ++ArgP;
                ShowImage();
                break;
            case XK_BackSpace:
                if (ArgP > 1)
                {
                    PrintPoints();
                    --ArgP;
                    ShowImage();
                }
                break;
            case XK_q:
                PrintPoints();
                return -1;
            default:
                break;
          }
          break;

      case MapNotify:
      case ReparentNotify:
          break;
      default:
          printf("Unknown event: %d\n", event.type);
    }
    return 0;
}

int main(int argc, char** argv)
{
    if (argc <= 1) {
        printf("Usage: blah blah\n");
        return 1;
    }
    ArgV = argv;
    ArgC = argc;
    ArgP = 1;

    if ((dpy = XOpenDisplay(getenv("DISPLAY"))) == 0)
    {
        fprintf(stderr, "Can't open display: %s\n", getenv("DISPLAY"));
        exit(1);
    }
    screen = DefaultScreen(dpy);

    win = XCreateSimpleWindow(dpy, RootWindow(dpy, screen),
                              0, 0, XSize, YSize, 3,
                              WhitePixel(dpy, screen),
                              BlackPixel(dpy, screen));
    if (!win)
    {
        fprintf(stderr, "Can't create window\n");
        exit(1);
    }

    gc = XCreateGC(dpy, win,  0, 0);
    if (!gc) {
        fprintf(stderr, "Couldn't create gc!\n");
        return 1;
    }

    gdk_pixbuf_xlib_init(dpy, screen);

    XSelectInput(dpy, win, eventMask);

    ShowImage();
    XMapWindow(dpy, win);

    while (HandleEvent() >= 0)
        ;
    return 0;
}

